var Immutable, assign, callUpdater, core, getNewStore, recorderEmit, recorderListeners;

assign = require("object-assign");

Immutable = require("immutable");

core = {
  records: Immutable.List(),
  pointer: 0,
  isTravelling: false,
  initial: Immutable.Map(),
  cachedStore: Immutable.Map(),
  updater: function(state) {
    return state;
  },
  inProduction: false
};

recorderListeners = Immutable.List();

recorderEmit = function(store, core) {
  return recorderListeners.forEach(function(fn) {
    return fn(store, core);
  });
};

callUpdater = function(actionType, actionData) {
  var backward, chunks, groupName, nextPointer, updater;
  chunks = actionType.split("/");
  groupName = chunks[0];
  updater = function(acc, action) {
    return core.updater(acc, action.get(0), action.get(1));
  };
  if (groupName === "actions-recorder") {
    switch (chunks[1]) {
      case "commit":
        return {
          initial: core.records.reduce(updater, core.initial),
          records: Immutable.List(),
          pointer: 0,
          isTravelling: false
        };
      case "reset":
        return {
          records: Immutable.List(),
          pointer: 0,
          isTravelling: false
        };
      case "peek":
        return {
          pointer: actionData,
          isTravelling: true
        };
      case "run":
        return {
          isTravelling: !core.isTravelling,
          pointer: 0
        };
      case "merge-before":
        if (core.pointer === 0) {
          return {};
        } else {
          return {
            initial: core.records.slice(0, core.pointer - 1).reduce(updater, core.initial),
            records: core.records.slice(core.pointer - 1),
            pointer: 1
          };
        }
        break;
      case "clear-after":
        return {
          records: core.records.slice(0, core.pointer)
        };
      case 'step':
        backward = actionData;
        if (backward) {
          if (core.pointer > 0) {
            nextPointer = core.pointer - 1;
          } else {
            nextPointer = core.pointer;
          }
        } else {
          if (core.pointer < core.records.size) {
            nextPointer = core.pointer + 1;
          } else {
            nextPointer = core.pointer;
          }
        }
        return {
          pointer: nextPointer
        };
      default:
        console.warn("Unknown actions-recorder action: " + actionType);
        return {};
    }
  } else {
    return {
      records: core.records.push(Immutable.List([actionType, actionData]))
    };
  }
};

getNewStore = function() {
  var updater;
  updater = function(acc, action) {
    return core.updater(acc, action.get(0), action.get(1));
  };
  if (core.isTravelling && core.pointer >= 0) {
    return core.records.slice(0, core.pointer).reduce(updater, core.initial);
  } else {
    return core.records.reduce(updater, core.initial);
  }
};

exports.setup = function(options) {
  assign(core, options);
  return core.cachedStore = core.initial;
};

exports.request = function(fn) {
  return fn(core.cachedStore, core);
};

exports.getState = function() {
  return core.cachedStore;
};

exports.getCore = function() {
  return core;
};

exports.subscribe = function(fn) {
  return recorderListeners = recorderListeners.unshift(fn);
};

exports.unsubscribe = function(fn) {
  return recorderListeners = recorderListeners.filterNot(function(listener) {
    return listener === fn;
  });
};

exports.dispatch = function(actionType, actionData) {
  actionData = Immutable.fromJS(actionData);
  if (core.inProduction) {
    core.initial = core.updater(core.initial, actionType, actionData);
    core.cachedStore = core.initial;
    recorderEmit(core.initial, core);
  } else {
    assign(core, callUpdater(actionType, actionData));
    core.cachedStore = getNewStore();
    recorderEmit(core.cachedStore, core);
  }
};
